-- PassMe CPLD code for patching Nintendo DS cartridge transfer allowing to run homebrew code.
-- PassMe is designed by Natrium42. http://www.dslinux.org/blogs/natrium42
-- Basic concept, CRC16 implementation and optimization by DarkFader. http://darkfader.net/ds/
-- 100% used of Xilinx XC9572XL-VQ44. Enable all area/density optimization settings.
-- ARM7 boots from address 0x080000C0.
-- Boot ARM9 with: *(volatile unsigned int *)0x027FFE24 = 0x02004000;

-- standard libraries
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Xilinx primitive components
library UNISIM;
use UNISIM.VComponents.all;

entity PassMe is
	port
	(
		DSSLOT_CLK		: in std_logic;
		DSSLOT_ROMCS	: in std_logic;
		DSSLOT_RESET	: in std_logic;
		DSSLOT_EEPCS	: in std_logic;
		DSSLOT_IRQ		: out std_logic;
		DSSLOT_IO		: inout std_logic_vector(7 downto 0);
		
		DSCART_CLK		: out std_logic;
		DSCART_ROMCS	: out std_logic;
		DSCART_RESET	: out std_logic;
		DSCART_EEPCS	: out std_logic;
		DSCART_IRQ 		: in std_logic;
		DSCART_IO 		: inout std_logic_vector(7 downto 0);
		
		--LED0 			: out std_logic;
		TX 			: out std_logic;
		RX				: in std_logic
	);
end PassMe;

architecture rtl of PassMe is
	signal is_command		: boolean;
	signal cmddata_cnt 		: natural range 0 to 511;		-- 8 + 504
	signal dsslot_clk_buf 	: std_logic;
	signal patched_data		: std_logic_vector(7 downto 0);
	signal next_crc16		: std_logic_vector(15 downto 0);
	signal crc16			: std_logic_vector(15 downto 0);
	signal patch_en			: boolean;
begin

	-- rs232
	TX <= '1' when DSSLOT_EEPCS = '1' else DSSLOT_IO(7);
	DSSLOT_IRQ <= (not RX) when (not patch_en) else '0';

	-- direct passthrough
	dsslot_clk_ibuf : IBUF port map ( I => DSSLOT_CLK, O => dsslot_clk_buf );
	dsslot_clk_obuf : OBUF port map ( I => dsslot_clk_buf, O => DSCART_CLK );
	DSCART_ROMCS <= DSSLOT_ROMCS;
	DSCART_RESET <= DSSLOT_RESET;
	--DSSLOT_IRQ <= DSCART_IRQ;
	DSCART_EEPCS <= DSSLOT_EEPCS;

	-- activity LED
	--LED0 <= not DSSLOT_ROMCS;

	-- CRC16
	next_crc16(15) <= patched_data(0) xor patched_data(1) xor patched_data(2) xor patched_data(3) xor patched_data(4) xor patched_data(5) xor patched_data(6) xor patched_data(7) xor crc16(7) xor crc16(6) xor crc16(5) xor crc16(4) xor crc16(3) xor crc16(2) xor crc16(1) xor crc16(0);
	next_crc16(14) <= patched_data(0) xor patched_data(1) xor patched_data(2) xor patched_data(3) xor patched_data(4) xor patched_data(5) xor patched_data(6) xor crc16(6) xor crc16(5) xor crc16(4) xor crc16(3) xor crc16(2) xor crc16(1) xor crc16(0);
	next_crc16(13) <= patched_data(6) xor patched_data(7) xor crc16(7) xor crc16(6);
	next_crc16(12) <= patched_data(5) xor patched_data(6) xor crc16(6) xor crc16(5);
	next_crc16(11) <= patched_data(4) xor patched_data(5) xor crc16(5) xor crc16(4);
	next_crc16(10) <= patched_data(3) xor patched_data(4) xor crc16(4) xor crc16(3);
	next_crc16(9) <= patched_data(2) xor patched_data(3) xor crc16(3) xor crc16(2);
	next_crc16(8) <= patched_data(1) xor patched_data(2) xor crc16(2) xor crc16(1);
	next_crc16(7) <= patched_data(0) xor patched_data(1) xor crc16(15) xor crc16(1) xor crc16(0);
	next_crc16(6) <= patched_data(0) xor crc16(14) xor crc16(0);
	next_crc16(5) <= crc16(13);
	next_crc16(4) <= crc16(12);
	next_crc16(3) <= crc16(11);
	next_crc16(2) <= crc16(10);
	next_crc16(1) <= crc16(9);
	next_crc16(0) <= patched_data(0) xor patched_data(1) xor patched_data(2) xor patched_data(3) xor patched_data(4) xor patched_data(5) xor patched_data(6) xor patched_data(7) xor crc16(8) xor crc16(7) xor crc16(6) xor crc16(5) xor crc16(4) xor crc16(3) xor crc16(2) xor crc16(1) xor crc16(0);

	-- dataswitcher
	process (DSSLOT_RESET, DSSLOT_ROMCS, DSSLOT_EEPCS, DSSLOT_IO, DSCART_IO, patched_data, crc16)
	begin
		DSSLOT_IO <= (others => 'Z');				-- default is high impedance
		DSCART_IO <= (others => 'Z');				-- default is high impedance
		patched_data <= DSCART_IO;

		if (DSSLOT_RESET='1') then					-- if not reset
			if (DSSLOT_ROMCS='0') then				-- ROM is selected
				if (is_command) then				-- is command byte
					DSCART_IO <= DSSLOT_IO;			-- from DS to cartridge
				else								-- is data byte
					DSSLOT_IO <= patched_data;		-- patched data to DS

					if (patch_en) then				-- patch enabled
						case cmddata_cnt is
							-- ldr pc, 0x027FFE24
							when 8+16#004# => patched_data <= X"18";
							when 8+16#005# => patched_data <= X"F0";
							when 8+16#006# => patched_data <= X"9F";
							when 8+16#007# => patched_data <= X"E5";

							-- set autostart bits
							when 8+16#01F# => patched_data <= X"04";
							when 8+16#022# => patched_data <= X"01";

							-- patch ARM9 entry address to endless loop
							when 8+16#024# => patched_data <= X"04";
							when 8+16#025# => patched_data <= X"FE";
							when 8+16#026# => patched_data <= X"7F";
							when 8+16#027# => patched_data <= X"02";

							-- patch ARM7 entry address
							when 8+16#034# => patched_data <= X"C0";
							when 8+16#035# => patched_data <= X"00";
							when 8+16#036# => patched_data <= X"00";
							when 8+16#037# => patched_data <= X"08";

							-- calculated CRC16
							when 8+16#15E# => patched_data <= crc16(7 downto 0);
							when 8+16#15F# => patched_data <= crc16(15 downto 8);

							when others => null;
						end case;
					end if;
				end if;
			elsif (DSSLOT_EEPCS='0') then			-- EEPROM is selected
				DSCART_IO(7) <= DSSLOT_IO(7);		-- pass serial data
				--DSSLOT_IO(6) <= DSCART_IO(6);		-- pass serial data in opposite direction
			 	DSSLOT_IO(6) <= RX;
			end if;
		end if;
	end process;

	-- patch_en
	process (DSSLOT_RESET, dsslot_clk_buf)
	begin
		if (DSSLOT_RESET='0') then
			patch_en <= true;						-- patch header
		elsif (rising_edge(dsslot_clk_buf)) then
			if (is_command) then
				if (DSCART_IO(5) = '1') then		-- detect 3C command, assume other command bytes are 00
					patch_en <= false;				-- do not patch other data
				end if;
			end if;
		end if;
	end process;

	-- crc16
	process (is_command, dsslot_clk_buf)
	begin
		if (is_command) then
			crc16 <= (others => '1');				-- init CRC16
		elsif (rising_edge(dsslot_clk_buf)) then
			if (cmddata_cnt /= 8+16#15E#) then		-- do not destroy high byte of header CRC
			    crc16 <= next_crc16;
			end if;
		end if;
	end process;

	-- cmddata_cnt, is_command
	process (DSSLOT_ROMCS, dsslot_clk_buf)
	begin
		if (DSSLOT_ROMCS='1') then
			cmddata_cnt <= 0;						-- new transfer
			is_command <= true;						-- start with command
		elsif (rising_edge(dsslot_clk_buf)) then
			if (cmddata_cnt = 7) then
				is_command <= false;				-- next byte is data
			end if;
			cmddata_cnt <= cmddata_cnt + 1;			-- next byte
		end if;
	end process;


end rtl;